#___MODULES___
from __future__ import annotations
from typing import Type
import datetime
import subprocess
import logging
import contextlib
from typing import Union

# pip installed modules
try:
        import pywinauto
        from pywinauto.application import process_get_modules
except ImportError: # requires pip install
        raise ModuleNotFoundError('To install the required modules use pip install pywinauto (Windows ONLY)')

try:
    from pyautogui import press
except ImportError: # requires pip install
        raise ModuleNotFoundError('To install the required modules use pip install pyautogui')


#___MY_MODULES___
from pywinauto.controls.uiawrapper import UIAWrapper
from pyuiauto.win.components import UIBaseComponent, UIWindow, UIButton, UIMenuItem, UIMenuBarItem
from pyuiauto.base.application import UIApplicationWrapper, UISystemTrayIconWrapper, UIPopupMenuWrapper
from pyuiauto.exceptions import ElementNotFound, WindowNotFound

class UIBackendExplorer():
    backend_explorer_app = pywinauto.Application(backend="uia").connect(path="explorer.exe")
    backend_explorer_app = pywinauto.Application(backend="uia").connect(path="explorer.exe")
    taskbar: pywinauto.WindowSpecification = backend_explorer_app.window(title="Taskbar", class_name="Shell_TrayWnd")

    try:
        taskbarExpand = UIButton(taskbar.child_window(title="Show Hidden Icons", class_name="SystemTray.NormalButton").wrapper_object())
    except:
        try:
            taskbarExpand = UIButton(taskbar.child_window(title="Notification Chevron").wrapper_object())
        except:
            raise ElementNotFound("Show Hidden Icons not found.")
        
backendExplorer = UIBackendExplorer()

class UISystemTrayIcon(UISystemTrayIconWrapper):
    def __init__(self, app: UIApplicationWrapper):
        super().__init__(app)

        self._iconHidden = False
        
        try:
            self._getIcon(backendExplorer.taskbar)
        except:
                self.__systrayExpandWindow = None
                self._iconHidden = True
    
    def _getIcon(self, parent: pywinauto.WindowSpecification) -> UIButton:
        icon = parent.child_window(title_re=f".* {self.app.appName}", control_type="Button", found_index=0)
        icon.wait('exists', timeout=2)
        icon.wrapper_object()
        return UIButton(icon)

    def __openSystemTrayExpand(self) -> pywinauto.WindowSpecification:
        # Open the system tray
        backendExplorer.taskbarExpand.invoke()

        # Select the audient icon
        try:
            systray = pywinauto.Application(backend="uia").connect(class_name="TopLevelWindowForOverflowXamlIsland")
            systrayWindow = systray.window(class_name="TopLevelWindowForOverflowXamlIsland")
            
        except:
            try:
                systray = pywinauto.Application(backend="uia").connect(class_name="NotifyIconOverflowWindow")
                systrayWindow = systray.window(class_name="NotifyIconOverflowWindow")
                
            except:
                raise WindowNotFound("Notification Overflow Window not found.")

        systrayWindow.wait("visible")
        return systrayWindow

    def __closeNotifictionExpand(self) -> None:
        # Close the system tray
        if self.__systrayExpandWindow.exists() and self.__systrayExpandWindow.is_visible():
            backendExplorer.taskbarExpand.invoke() # pyautogui.press("esc") will also work for closing popup menu windows
            self.__systrayExpandWindow.wait_not("visible")

    def __enter__(self) -> UIButton:
        if self._iconHidden:
            try:
                self.__systrayExpandWindow = self.__openSystemTrayExpand()
                return self._getIcon(self.__systrayExpandWindow)
            except: raise ElementNotFound("System Tray Icon not found.")
        
        return self._getIcon(backendExplorer.taskbar)
    
    def __exit__(self, *args):
        if self._iconHidden:
            self.__closeNotifictionExpand()


class UIPopupMenu(UIPopupMenuWrapper):
    def __init__(self, app: UIApplicationWrapper, popup_naming_scheme: str = None) -> None:
        super().__init__(app, popup_naming_scheme)

    def getMenuItemFromPath(self, *path: str) -> UIMenuItem:
        current_item = None

        for count, i in enumerate(path):
            self.current_popup = self.app.window(title=self.win_name, found_index=0)
            current_item = self.current_popup.findFirstR(title=i, control_type=UIMenuItem)
            if count != len(path)-1:
                current_item.invoke()
            self.steps += 1
        
        return current_item
    
    def back(self):
        if self.steps > 0:
            press("left")
            self.steps -= 1

    def backToRoot(self):
        for i in range(self.steps):
            press("left")
            self.steps -= 1
    
    def __enter__(self) -> UIPopupMenu:
        return self
        
    def __exit__(self, *args) -> None:
        if self.current_popup.isVisible():
            press("esc")


#___DEFINING_NATIVE_METHODS___
class UIApplication(UIApplicationWrapper):

    def launchApp(self) -> None:
        pywinauto.Application().start(self.appPath)
    
    def connectApp(self) -> None:
        self._app = pywinauto.Application(backend="uia", allow_magic_lookup=False).connect(title=self.appName, path=self.appPath)
    
    def terminateApp(self) -> None:
        self._app.kill()
    
    def window(self, timeout: int = 1, retry_interval: float = 0.01, **criteria) -> UIWindow:
        try:
            window = self._app.window(**criteria)
            window.wait('exists', timeout=timeout, retry_interval=retry_interval)
            return UIWindow(window.wrapper_object())
        except pywinauto.timings.TimeoutError:
             raise WindowNotFound
    
    def windows(self, timeout: int = 1, retry_interval: float = 0.01, **criteria) -> list[UIWindow]:
        try:
            windows = list(UIWindow(window) for window in self._app.windows(**criteria))
            if len(windows) == 0:
                raise WindowNotFound
            return windows
        except pywinauto.timings.TimeoutError:
             raise WindowNotFound
    
    def isAppAlreadyRunning(self):
        for module in process_get_modules():
            if self.appPath == module[1]:
                 return True
        return False

    def isAppRunning(self):
        return self._app.is_process_running()
    
    def getCrashReport(self):
        # Convert epoch timestamp to a datetime object
        start_datetime = datetime.datetime.fromtimestamp(self.start_time)
        end_datetime = datetime.datetime.fromtimestamp(self.end_time)

        # Format datetime object as a string in the required format
        start_time = start_datetime.strftime('%Y-%m-%dT%H:%M:%S')
        end_time = end_datetime.strftime('%Y-%m-%dT%H:%M:%S')

        command = f"powershell.exe Get-WinEvent -FilterHashtable @{{Logname='Application';Data='{self.appName}';StartTime='{start_time}';EndTime='{end_time}'}}"
        result = subprocess.run(command, capture_output=True, text=True)

        if result.returncode == 0:
            return result.stdout
        else:
            return result.stderr
    
    def getPopupMenu(self, popup_naming_scheme: str = None) -> UIPopupMenu:
        return UIPopupMenu(self, popup_naming_scheme)
    
    def getSystemTrayIcon(self) -> UISystemTrayIcon:
        return UISystemTrayIcon(self)

    @contextlib.contextmanager
    def systemTrayPopupPath(self, *path: str) -> UIMenuItem:
        with self.getSystemTrayIcon() as icon:
            icon.right_click()
        
            with self.getPopupMenu() as popup:
                try:

                    yield popup.getMenuItemFromPath(*path)
                except ElementNotFound:
                    raise ElementNotFound(f"{path} is disabled or not available")
    
    @contextlib.contextmanager
    def menuBarPopupPath(self, window: UIWindow, *path: str) -> Union[UIMenuBarItem, UIMenuItem]:
        try:
            menuBarItem = window.findFirstR(title=path[0], control_type=UIMenuBarItem)
            if len(path) == 1:
                yield menuBarItem
            
            menuBarItem.click()
            with self.getPopupMenu() as popup:
                yield popup.getMenuItemFromPath(*path[1:])
        
        except ElementNotFound:
                    raise ElementNotFound(f"{path} is disabled or not available")



# ================== Component Exceptions ==================

class InvalidElement(RuntimeError):
    """Raises when an invalid element is passed"""
    pass


class ElementNotVisible(RuntimeError):
    """Raised when an element is not visible"""
    pass


class ElementNotFound(RuntimeError):
    """Raised when an element is not found"""
    pass

class WindowNotFound(RuntimeError):
    """Raised when a window is not found"""
    pass

# ================= Application Exceptions =================

class ProcessNotFoundError(Exception):
    """Could not find that process"""
    pass


class AppNotConnected(Exception):
    """Application has not been connected to a process yet"""
    pass

# ==========================================================



import time

def wait_until_condition(conditional_func, timeout: float = 3, retry_interval: float = 0.01, *func_args, **func_kwargs) -> bool:
    '''Wait until condition function\n
    Returns True if the conditional function returns a value that is True before the Timeout is reached.\n
    However, returns False if the conditional function doesn't return a value that is True before the Timeout is reached.\n
    Args:
                    conditional_func: the function to perform on each iteration check
                    
                    (optional)
                    timeout: the timeout max (in seconds) of type float
                    retry_interval: the delay time (in seconds) of type float between each check
                    *func_args: the arguments to pass into the conditional function
                    **func_kwargs: the keyword arguments to pass into the conditional function

        Returns: 
                    result: value of type bool
    '''
    timeafter = time.time() + timeout
    while time.time() < timeafter:
        if conditional_func(*func_args, **func_kwargs): return True
        time.sleep(retry_interval)
    return False

def wait_until_raise(conditional_func, error: Exception = TimeoutError(), timeout: float = 3, retry_interval: float = 0.01, *func_args, **func_kwargs):
    '''Wait until raise function\n
    Returns the value if the conditional function returns a value that is True before the Timeout is reached.\n
    However, raises an error if the conditional function doesn't return a value that is True before the Timeout is reached.\n
    Args:
                    conditional_func: the function to perform on each iteration check
                    
                    (optional)
                    error: the error to raise if a timeout is reached of type exception
                    timeout: the timeout max (in seconds) of type float
                    retry_interval: the delay time (in seconds) of type float between each check
                    *func_args: the arguments to pass into the conditional function
                    **func_kwargs: the keyword arguments to pass into the conditional function

        Returns: 
                    result: value of type bool
    '''
    timeafter = time.time() + timeout
    while time.time() < timeafter:
        result = conditional_func(*func_args, **func_kwargs)
        if result: return result
        time.sleep(retry_interval)
    raise error


# ___MODULES___
from __future__ import annotations
from abc import ABC, ABCMeta, abstractmethod
from typing import Type, Union

# ___CLASSES___

class UIBaseComponentWrapperMeta(ABCMeta):
    """Metaclass for UIComponentWrapper objects"""

    control_type_to_cls = {}

    def __init__(cls, name, bases, attrs):
        """Register the control types"""

        if name != "UIBaseComponentWrapper":
            UIBaseComponentWrapperMeta.control_type_to_cls[cls.native_control_type] = cls

    @staticmethod
    def find_wrapper(component):
        """Find the correct wrapper for this UIA component"""
        raise NotImplementedError

class UIBaseComponentWrapper(object, metaclass=UIBaseComponentWrapperMeta):
    """
    Abstract / Default wrapper for User Interface controls.

    All other UI wrappers are derived from this.

    This class wraps a lot of functionality of underlying UI features
    for working with Mac or Windows.

    Most of the methods apply to every single component type. For example
    you can click() on any component.

    """
    
    def __new__(cls, component):
        return UIBaseComponentWrapper._create_wrapper(cls, component, UIBaseComponentWrapper)
    
    @staticmethod
    def _create_wrapper(cls_spec, component, myself):
        """Create a wrapper object according to the specified component"""
        # only use the meta class to find the wrapper for BaseWrapper
        # so allow users to force the wrapper if they want
        
        if cls_spec != myself:
            obj = object.__new__(cls_spec)
            obj.__init__(component)
            return obj

        new_class = cls_spec.find_wrapper(component)
        
        obj = object.__new__(new_class)

        obj.__init__(component)

        return obj

    def __init__(self, component):
        """
        Initialize the component
        """
        if component:
            
            self.component = component
        else:
            raise RuntimeError('NULL pointer was used to initialize BaseWrapper')
    
    @abstractmethod
    def setValue(self, value: Union[bool, str, int, float]):
        'Standard base class set value method (may not currently work on all components).'
        raise NotImplementedError

    @abstractmethod
    def getValue(self) -> Union[bool, str, int, float]:
        'Standard base class get value method (may not currently work on all components).'
        raise NotImplementedError
    
    @abstractmethod
    def setFocus(self):
        '''Standard base class set focus method (may not currently work on all components).\n
        Brings the components root parent window to the top most level (z-index) and sets component as "active".'''
        raise NotImplementedError

    @abstractmethod
    def invoke(self):
        '''Standard base class invoke method (may not currently work on all components).\n
        Invoke will select / click the component without actual mouse movement
        (therefore this may work on components that are not visible).'''
        raise NotImplementedError

    @abstractmethod
    def click(self):
        '''Standard base class click method\n
        Manually performs a mouse movement and clicks the component at it's mid-point.'''
        raise NotImplementedError

    @abstractmethod
    def right_click(self):
        '''Standard base class right click method\n
        Manually performs a mouse movement and right clicks the component at it's mid-point.'''
        raise NotImplementedError
    
    @abstractmethod
    def isTopLevel(self) -> bool:
        '''Standard base class is on top method\n
        Compares the components position against the top most level (z-index) component at this position.'''
        raise NotImplementedError

    @abstractmethod
    def isVisible(self) -> bool:
        '''Standard base class is visible method\n
        Uses a OS dependant accessibility API checks.'''
        raise NotImplementedError

    @abstractmethod
    def exists(self) -> bool:
        '''Standard base class exists method\n
        Uses a OS dependant accessibility API checks.'''
        raise NotImplementedError
    
    @abstractmethod
    def getMidpoint(self) -> tuple[int, int]:
        '''Standard base class get mid point method\n
        Finds the components mid-point.\n
        Returns: 
                    x: x-position in pixels of type int
                    y: y-position in pixels of type int'''
        raise NotImplementedError
    
    @abstractmethod
    def getCoordinates(self) -> tuple[int, int, int, int]:
        '''Standard base class get coordinates method\n
        Gets the components rectangular edge positions.\n
        Returns: 
                    left: x-position of left top corner in pixels of type int
                    top: y-position of left top corner in pixels of type int
                    right: x-position of bottom right corner in pixels of type int
                    bottom: y-position of bottom right corner in pixels of type int'''
        raise NotImplementedError
    
    @abstractmethod
    def getSizes(self):
        '''Standard base class get sizes method\n
        Gets the components width and height.\n
        Returns: 
                    width: width in pixels of type int
                    height: height in pixels of type int'''
        raise NotImplementedError

    @abstractmethod
    def getChildren(self) -> UIBaseComponentWrapper:
        '''Standard base class get children method\n
        Gets all of the components children (non recursive).\n
        Returns: 
                    children: all children in a list of type UIBaseComponentWrapper
                    (type should dynamicly wrap to it's cross compatible wrapper)'''
        raise NotImplementedError
    
    @abstractmethod
    def getDescendants(self) -> UIBaseComponentWrapper:
        '''Standard base class get descendants method\n
        Gets all of the components descendants (recursive).\n
        Returns: 
                    descendants: all descendants in a list of type UIBaseComponentWrapper
                    (type should dynamicly wrap to it's cross compatible wrapper)'''
        raise NotImplementedError

    @abstractmethod
    def findAll(self, control_type: Type, timeout: int = 1, retry_interval: float = 0.01, **criteria) -> list[UIBaseComponentWrapper]:
        '''Standard base class find all method\n
        Gets all of the components children with the specified criteria.
        Raises an exception if the criteria matches no components (pre order traversal, non recursive).\n
        Args:
                    control_type: the specified control type wrapper to search (of type UIBaseComponentWrapper)
                    
                    (optional criteria)
                    title: the name of the component (of type string)

                    (extra OS specific accessibility API criteria)
        Returns: 
                    components: components dynamicly wrapped to it's cross compatible custom wrapper in a list of type UIBaseComponentWrapper'''
        raise NotImplementedError
    
    @abstractmethod
    def findFirst(self, control_type: Type, timeout: int = 1, retry_interval: float = 0.01, **criteria) -> UIBaseComponentWrapper:
        '''Standard base class find first method\n
        Gets the first child component with the specified criteria.
        Raises an exception if the criteria matches no components (pre order traversal, non recursive).\n
        Args:
                    control_type: the specified control type wrapper to search (of type UIBaseComponentWrapper)
                    
                    (optional criteria)
                    title: the name of the component (of type string)

                    (extra OS specific accessibility API criteria)
        Returns: 
                    component: component dynamicly wrapped to it's cross compatible custom wrapper of type UIBaseComponentWrapper'''
        raise NotImplementedError
    
    @abstractmethod
    def findAllR(self, control_type: Type, timeout: int = 1, retry_interval: float = 0.01, **criteria) -> list[UIBaseComponentWrapper]:
        '''Standard base class find all recursive method\n
        Gets all of the components children with the specified criteria.
        Raises an exception if the criteria matches no components (pre order traversal, recursive).\n
        Args:
                    control_type: the specified control type wrapper to search (of type UIBaseComponentWrapper)
                    
                    (optional criteria)
                    title: the name of the component (of type string)

                    (extra OS specific accessibility API criteria)
        Returns: 
                    components: components dynamicly wrapped to it's cross compatible custom wrapper in a list of type UIBaseComponentWrapper'''
        raise NotImplementedError
    
    @abstractmethod
    def findFirstR(self, control_type: Type, timeout: int = 1, retry_interval: float = 0.01, **criteria) -> UIBaseComponentWrapper:
        '''Standard base class find first recursive method\n
        Gets the first child component with the specified criteria.
        Raises an exception if the criteria matches no components (pre order traversal, recursive).\n
        Args:
                    control_type: the specified control type wrapper to search (of type UIBaseComponentWrapper)
                    
                    (optional criteria)
                    title: the name of the component (of type string)

                    (extra OS specific accessibility API criteria)
        Returns: 
                    component: component dynamicly wrapped to it's cross compatible custom wrapper of type UIBaseComponentWrapper'''
        raise NotImplementedError
    
    @abstractmethod
    def find(self, control_type: Type, timeout: int = 1, retry_interval: float = 0.01, **criteria) -> UIBaseComponentWrapper:
        '''Standard base class find method\n
        Finds a child component with the specified criteria.
        Raises an exception if the criteria matches > 1 or no components (non recursive).\n
        Args:
                    control_type: the specified control type wrapper to search (of type UIBaseComponentWrapper)
                    
                    (optional criteria)
                    title: the name of the component (of type string)

                    (extra OS specific accessibility API criteria)
        Returns: 
                    component: component dynamicly wrapped to it's cross compatible custom wrapper of type UIBaseComponentWrapper'''
        raise NotImplementedError

    @abstractmethod
    def findR(self, control_type: Type, timeout: int = 1, retry_interval: float = 0.01, **criteria) -> UIBaseComponentWrapper:
        '''Standard base class find recursive method\n
        Finds a child component with the specified criteria.
        Raises an exception if the criteria matches > 1 or no components (recursive).\n
        Args:
                    control_type: the specified control type wrapper to search (of type UIBaseComponentWrapper)
                    
                    (optional criteria)
                    title: the name of the component (of type string)

                    (extra OS specific accessibility API criteria)
        Returns: 
                    component: component dynamicly wrapped to it's cross compatible custom wrapper of type UIBaseComponentWrapper'''
        raise NotImplementedError
    
    @classmethod
    @property
    @abstractmethod
    def native_control_type(cls) -> str:
        'Standard base class control type property (name of the components native control type).'
        raise NotImplementedError
    
    @classmethod
    @property
    @abstractmethod
    def control_type(cls) -> Type:
        'Standard base class control type property (name of the components control type).'
        raise NotImplementedError
    
    @property
    @abstractmethod
    def title(self) -> str:
        'Standard base class title property (name of the component).'
        raise NotImplementedError

    # _protected methods

    def _check_element_exists(self, component) -> bool:
        return not(component is None)

    def _check_elements_exist(self, components) -> bool:
        return len(components) > 0
    
    # def _wrapper_function_from_native(self, component, native_control_type: str) -> UIBaseComponentWrapper:
    #     '''Wrap objects using their native component type
    #     '''
    #     return wrapper_mapping[native_control_type](component)
    
    # def _wrapper_function(self, component, control_type: Type) -> UIBaseComponentWrapper:
    #     '''Wrap objects using the pyUIauto control types\n
    #     ( also checks if the elements aren't null )
    #     '''
    #     return control_type(component) if self._check_element_exists(component) else None
    
    # def _list_wrapper_function(self, components, control_type: Type) -> list[UIBaseComponentWrapper]:
    #     '''Wrap objects using the pyUIauto control types\n
    #     ( also checks if the elements aren't null )
    #     '''
    #     return list(control_type(component) for component in components) if self._check_elements_exist(components) else None

    def __str__(self) -> str:
        return self.control_type

class UIButtonWrapper():
    @abstractmethod
    def press(self):
        '''Button press method\n
        Similar to the invoke method, performs a button press without actual mouse movement.
        (therefore this may work on components that are not visible).'''
    
    def toggle(self, value):
        '''Button toggle method\n
        Performs a button press if the provided value != current value.
        (overrides the setValue method)'''
        return self.press() if self.getValue() != value else None

    def setValue(self, value):
        return self.toggle(value)

class UIRadioButtonWrapper(): ...

class UITextWrapper(): ...

class UISliderWrapper(): ...

class UIEditWrapper(): ...

class UIMenuWrapper(): ...
class UIMenuItemWrapper():
    @abstractmethod
    def select(self):
        '''Menu item select method\n
        Similar to the invoke method, selects a menu item without actual mouse movement.'''
    
    def invoke(self):
        return self.select()

    def toggle(self, value) -> bool:
        '''Menu item toggle method\n
        Selects a menu item if the provided value != current value.
        (overrides the setValue method)'''
        if self.getValue() != value:
            self.select()
            return True
        else:
            return False
    
    def setValue(self, value):
        return self.toggle(value)

class UIWindowWrapper():
    @abstractmethod
    def moveResize(self, x: int = None, y: int = None, width: int = None, height: int = None):
        '''Window move resize method\n
        Moves a windows top left corner to the specified pixel points x, y (if provided)
        and resizes the window to the specified pixel width, height (if provided)
        Args:
                    (optional | default = None)
                    x: top left corner x-axis in pixels (of type int)
                    y: top left corner y axis in pixels (of type int)
                    width: width of the window size in pixels (of type int)
                    height: height of the window size in pixels (of type int)'''

    @abstractmethod
    def close(self):
        '''Window close method to press the close button\n
        Similar to a button invoke method, selects the window's close button without actual mouse movement.'''

class UIGroupWrapper(): ...
class UIStaticTextWrapper(): ...

class UIMenuBarWrapper(): ...

class UIProgressBarWrapper(): ...

class UITitleBarWrapper(): ...

# !!! Mac Specific Components !!!
class UIMenuBarItemWrapper(): ...


# wrapper_mapping = {wrapper.native_control_type: wrapper for wrapper in (UIButtonWrapper, UIRadioButtonWrapper, UIMenuWrapper, UIMenuItemWrapper, UIProgressBarWrapper, UITextWrapper, UISliderWrapper, UIEditWrapper, UITitleBarWrapper, UIMenuBarWrapper, UIMenuBarItemWrapper, UIWindowWrapper, UIGroupWrapper, UIStaticTextWrapper)}



from platform import system

if system() == "Darwin":
    from pyuiauto.mac.components import UIButton, UIEdit, UIMenu, UIGroup, UIMenuBar, UIMenuBarItem, UIMenuItem, UISlider, UIProgressBar, UIRadioButton, UIWindow, UIText, UIStaticText, UITitleBar

elif system() == "Windows":
    from pyuiauto.win.components import UIButton, UIEdit, UIMenu, UIGroup, UIMenuBar, UIMenuBarItem, UIMenuItem, UISlider, UIProgressBar, UIRadioButton, UIWindow, UIText, UIStaticText, UITitleBar


#___MODULES___
from __future__ import annotations
from typing import Type
import time
from abc import ABCMeta
import logging

# pip installed modules
try:
    from atomacos import NativeUIElement
except ImportError: # requires pip install
    raise ModuleNotFoundError('To install the required modules use pip install atomacos (MacOS ONLY)')


from ApplicationServices import AXValueCreate, kAXValueCGPointType, kAXValueCGSizeType
import Quartz.CoreGraphics as CG

#___MY_MODULES___
from pyuiauto.exceptions import ElementNotFound
from pyuiauto.base.components import *


#___DEFINING_NATIVE_WRAPPER_META___

class UIBaseComponentMeta(UIBaseComponentWrapperMeta):
    """Metaclass for UIComponent objects"""

    control_type_to_cls = {}

    def __init__(cls, name, bases, attrs):
        """Register the control types"""

        UIBaseComponentWrapperMeta.__init__(cls, name, bases, attrs)

        if name != "UIBaseComponent":
            UIBaseComponentMeta.control_type_to_cls[cls.native_control_type] = cls

    @staticmethod
    def find_wrapper(component):
        """Find the correct wrapper for this UIA component"""
        # Set a general wrapper by default
        wrapper_match = UIBaseComponent

        # Check for a more specific wrapper in the registry
        try:
            control_type = component.AXRole
        except:
            control_type = "No Role"
            logging.warning(f"Component: {component} | has no control type / AXRole attribute")

        if control_type in UIBaseComponentWrapperMeta.control_type_to_cls:
            wrapper_match = UIBaseComponentWrapperMeta.control_type_to_cls[control_type]
        else:
            raise NotImplementedError(f"{component} doesn't have an implemented wrapper")

        return wrapper_match

#___DEFINING_NATIVE_METHODS___

class UIBaseComponent(UIBaseComponentWrapper, metaclass=UIBaseComponentMeta):

    def __new__(cls, component: NativeUIElement):
        """Construct the control wrapper"""
        return super(UIBaseComponent, cls)._create_wrapper(cls, component, UIBaseComponent)

    # -----------------------------------------------------------
    def __init__(self, component: NativeUIElement):
        UIBaseComponentWrapper.__init__(self, component)

    def getValue(self):
        return self.component.AXValue
    
    def setValue(self, value):
        self.component.AXValue = value
    
    def setFocus(self):
        self.component.activate()
    
    def invoke(self):
        self.component.Press()

    def click(self):
        self.component.clickMouseButtonLeft(self.getMidpoint())
    
    def right_click(self):
        self.component.clickMouseButtonRight(self.getMidpoint())

    def isTopLevel(self):
        x, y, _, _= self.getCoordinates()
        is_visible = self.component.getApplication().getElementAtPosition((x, y)) == self.component
        return is_visible

    def isVisible(self):
        x, y, _, _= self.getCoordinates()
        is_visible = self.component.getApplication().getElementAtPosition((x, y)) == self.component
        return is_visible

    def exists(self):
        return hasattr(self.component, "AXRole")
    
    def getMidpoint(self):
        return ((self.component.AXPosition[0] + self.component.AXSize[0] / 2), (self.component.AXPosition[1] + self.component.AXSize[1] / 2))

    def getCoordinates(self):
        x, y = self.component.AXPosition
        w, h = self.component.AXSize
        return x, y, x+w, y+h
    
    def getSizes(self):
        return self.component.AXSize
    
    def getChildren(self):
        return list(UIBaseComponent(component, component.AXRole) for component in self.component.AXChildren)
    
    getDescendants = getChildren
    
    def _wait_find(self, function, check_function, control_type: Type, timeout: int = 1, retry_interval: float = 0.01, **criteria):
        for i in (  {"conversion": "title", "apple": "AXTitle"},
                    {"conversion": "description", "apple": "AXDescription"},
                ):
            try:
                criteria[i["apple"]] = criteria.pop(i["conversion"])
            except KeyError:
                pass

        timeafter = time.time() + timeout
        while time.time() < timeafter:
            item = function(AXRole=control_type.native_control_type, **criteria)
            item_check = check_function(item)
            if item_check:
                if type(item) == list:
                    return list(UIBaseComponent(i) for i in item)
                else:
                    return UIBaseComponent(item)
            time.sleep(retry_interval)
        raise ElementNotFound(f"Element - ControlType: {control_type.native_control_type} - {criteria} - not found after {timeout} seconds")

    def findFirst(self, control_type: Type, timeout: int = 1, retry_interval: float = 0.01, **criteria):
        return self._wait_find(function=self.component.findFirst,
                                check_function=self._check_element_exists,
                                control_type=control_type,
                                timeout=timeout,
                                retry_interval = retry_interval,
                                **criteria)

    def findAll(self, control_type: Type, timeout: int = 1, retry_interval: float = 0.01, **criteria):
        return self._wait_find(function=self.component.findAll,
                                check_function=self._check_elements_exist,
                                control_type=control_type,
                                timeout=timeout,
                                retry_interval = retry_interval,
                                **criteria)

    def findFirstR(self, control_type: Type, timeout: int = 1, retry_interval: float = 0.01, **criteria):
        return self._wait_find(function=self.component.findFirstR,
                                check_function=self._check_element_exists,
                                control_type=control_type,
                                timeout=timeout,
                                retry_interval = retry_interval,
                                **criteria)

    def findAllR(self, control_type: Type, timeout: int = 1, retry_interval: float = 0.01, **criteria):
        return self._wait_find(function=self.component.findAllR,
                                check_function=self._check_elements_exist,
                                control_type=control_type,
                                timeout=timeout,
                                retry_interval = retry_interval,
                                **criteria)
    
    def find(self, control_type: Type, timeout: int = 1, retry_interval: float = 0.01, **criteria):
        items = self.findAll(control_type, timeout, retry_interval, **criteria)
        
        if len(items) > 1:
            raise ElementNotFound(f"{len(items)} Elements found matching the criteria - ControlType: {control_type} - {criteria}")

        return items[0]
    
    def findR(self, control_type: Type, timeout: int = 1, retry_interval: float = 0.01, **criteria):
        items = self.findAllR(control_type, timeout, retry_interval, **criteria)
        
        if len(items) > 1:
            raise ElementNotFound(f"{len(items)} Elements found matching the criteria - ControlType: {control_type} - {criteria}")

        return items[0]
    
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return None
    
    @classmethod
    @property
    def control_type(cls) -> Type:
        return type(cls)
    
    @property
    def title(self) -> str:
        return self.component.AXTitle


# ============================================


class UIAppRoot(UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "AXApplication"


# ============================================


class UIButton(UIButtonWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "AXButton"
    
    def press(self):
        self.invoke()
    
    def setValue(self, value):
        return super().setValue(value)

class UIRadioButton(UIRadioButtonWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "AXRadioButton"

class UIText(UITextWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "AXText"

class UISlider(UISliderWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "AXSlider"

class UIEdit(UIEditWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "AXTextArea"

class UIMenu(UIMenuWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "AXMenu"

class UIMenuItem(UIMenuItemWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "AXMenuItem"

    def getValue(self) -> bool:
        return self.component.AXMenuItemMarkChar != None

    def select(self):
        self.component.Press()
    
    def setValue(self, value):
        return super().setValue(value)

class UIWindow(UIWindowWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "AXWindow"

    def moveResize(self, x=None, y=None, width=None, height=None):
        left, top, _, _ = self.getCoordinates()
        w, h = self.getSizes()

        # if no X is specified - so use current coordinate
        if x is None:
            x = left

        # if no Y is specified - so use current coordinate
        if y is None:
            y = top

        # if no width is specified - so use current width
        if width is None:
            width = w

        # if no height is specified - so use current height
        if height is None:
            height = h

        # ask for the window to be moved
        point = CG.CGPoint(x=x, y=y)
        self.component.AXPosition = AXValueCreate(kAXValueCGPointType, point)

        # ask for the window to be resized
        size = CG.CGSize(width=width, height=height)
        self.component.AXSize = AXValueCreate(kAXValueCGSizeType, size)

    def close(self):
        self.component.AXCloseButton.Press()

class UIGroup(UIGroupWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "AXGroup"

class UIStaticText(UIStaticTextWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "AXStaticText"

class UITitleBar(UITitleBarWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "AXTitleBar"

class UIMenuBar(UIMenuBarWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "AXMenuBar"

class UIProgressBar(UIProgressBarWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "AXProgressIndicator"

# !!! Mac Specific !!!

class UIMenuBarItem(UIMenuItemWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "AXMenuBarItem"


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyuiauto', 'pyuiauto.base', 'pyuiauto.mac', 'pyuiauto.win']

package_data = \
{'': ['*']}

install_requires = \
['importlib-metadata>=6.6.0,<7.0.0',
 'pathlib>=1.0.1,<2.0.0',
 'pyautogui==0.9.41']

extras_require = \
{':sys_platform == "darwin"': ['atomacos>=3.3.0,<4.0.0'],
 ':sys_platform == "win32"': ['pywinauto>=0.6.8,<0.7.0']}

setup_kwargs = {
    'name': 'pyuiauto',
    'version': '0.1.7',
    'description': 'Python UI Automation library, for cross-platform applications, interfacing through the accessibility API',
    'long_description': '# pyUIauto\n\n[![macOS](https://svgshare.com/i/ZjP.svg)](https://svgshare.com/i/ZjP.svg)\n[![Windows](https://svgshare.com/i/ZhY.svg)](https://svgshare.com/i/ZhY.svg)\n[![PyPi version](https://badgen.net/pypi/v/pyuiauto/)](https://pypi.org/project/pyuiauto/)\n[![PyPi license](https://badgen.net/pypi/license/pyuiauto/)](https://pypi.org/project/pyuiauto/)\n[![PyPI pyversions](https://img.shields.io/pypi/pyversions/pyuiauto.svg)](https://pypi.python.org/pypi/pyuiauto/)\n\n\n| Tests       | Status                                                                                                                  |\n| :---------- | :---------------------------------------------------------------------------------------------------------------------: |\n| Development | ![Development Tests](https://github.com/harveyf2801/pyUIauto/actions/workflows/run_dev_tests.yml/badge.svg?branch=main)       |\n| Build       | ![Build Tests](https://github.com/harveyf2801/pyUIauto/actions/workflows/build_wheel.yml/badge.svg?branch=main) |\n\nPython UI Automation library, for cross-platform applications, interfacing through the accessibility API.\n\n## Description\n\nThis library / framework takes two popular UI automation libraries and combines their functionality by wrapping them into custom components and creating methods that function in similar ways for both OS. This project was originally designed as part of a QA automation project to perform end-to-end testing on desktop applications.\n\n## Getting Started\n\n### Dependencies\n\nPython Packages:\n\n- pywinauto (Windows / Linux)\n- atomacos (MacOS)\n- pyautogui\n\nOS Compatibility:\n\n- Windows\n- MacOS\n\n( Currently untested on Linux )\n\n## Example\n\n```python\n# Import the tools needed\nfrom platform import system\nimport os\nfrom pyuiauto.application import UIApplication\nfrom pyuiauto.components import UIButton\n\n# Finding the path location of the application\napp_paths = {\n  "Darwin": "/Applications/Visual Studio Code.app",\n  "Windows": os.path.expanduser(\'~\') + "\\\\AppData\\\\Local\\\\Programs\\\\Microsoft VS Code\\\\Code.exe"\n}\n\nif system() in app_paths:\n  appPath = app_paths[system()]\nelse:\n  raise NotImplementedError("The current OS is not currently supported: " + system())\n\n# Setting up an application template, launching the app, and connecting to it\napp = UIApplication(appName = "Visual Studio Code", appPath = appPath)\napp.launchApp()\napp.connectApp()\n\n# Finding the window component and searching for elements within this window component\nmain_window = app.window(title = "Visual Studio Code", timeout = 2)\nmain_window.findR(title = "Toggle Primary Side Bar (Ctrl+B)", control_type = UIButton).press() \'\'\'  press will invoke a button without manually moving the mouse and clicking it \n                                                                                          (a button could be invoked even if it isn\'t currently visible)  \'\'\'\nmain_window.findR(title = "Open Folder", control_type = UIButton).click() \'\'\' however, click will move the mouse to the button location and click it\n                                                                    (sometimes this can be more reliable) \'\'\'\n\n# Closing the window and terminating the application\nmain_window.close()\napp.terminateApp()\n```\n\n## Authors\n\nex. Harvey Fretwell\n\nex. [pyWinAuto](https://github.com/pywinauto/pywinauto/tree/master)\n\nex. [atomacos](https://github.com/daveenguyen/atomacos)\n\nex. [pyAutoGUI](https://github.com/asweigart/pyautogui)\n\n## Version History\n\n- 0.1\n  - Initial Release\n  - 0.1.1\n    - Added UISystemTrayIcon and UIPopupMenu manager\n  - 0.1.4\n    - Fixed some issues with setValue() method on buttons and menus\n  - 0.1.5\n    - Added context managers for better popup menu handling\n  - 0.1.6\n    - Added isOnTop method for components\n    - Fixed isVisible method for components\n    - Added checks for components existing in helper methods\n  - 0.1.7\n    - Added MacOS compatibility\n    - Fixed getValue method for mac on menu items (with a bit of a work around ... hoping to find a better solution soon)\n    - Fixed menu path select and system tray popup select methods on mac\n\n## Acknowledgments\n\n- [pyWinAuto](https://github.com/pywinauto/pywinauto/tree/master)\n- [atomacos](https://github.com/daveenguyen/atomacos)\n- [pyAutoGUI](https://github.com/asweigart/pyautogui)\n',
    'author': 'Harvey Fretwell',
    'author_email': 'hgfretwell@gmail.com',
    'maintainer': 'Harvey Fretwell',
    'maintainer_email': 'hgfretwell@gmail.com',
    'url': 'https://github.com/harveyf2801/pyUIauto',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



#___MODULES___
from __future__ import annotations
from typing import Type
import time
from abc import ABCMeta
import logging

# pip installed modules
try:
        from pywinauto.controls.uiawrapper import UIAWrapper
        from pywinauto.keyboard import send_keys
        from pywinauto.timings import wait_until
        from pywinauto.uia_defines import NoPatternInterfaceError
except ImportError: # requires pip install
        raise ModuleNotFoundError('To install the required modules use pip install pywinauto (Windows ONLY)')


#___MY_MODULES___
from pyuiauto.exceptions import ElementNotFound
from pyuiauto.base.components import *


#___DEFINING_NATIVE_WRAPPER_META___

class UIBaseComponentMeta(UIBaseComponentWrapperMeta):
    """Metaclass for UIComponent objects"""

    control_type_to_cls = {}

    def __init__(cls, name, bases, attrs):
        """Register the control types"""

        UIBaseComponentWrapperMeta.__init__(cls, name, bases, attrs)

        if name != "UIBaseComponent":
            UIBaseComponentMeta.control_type_to_cls[cls.native_control_type] = cls

    @staticmethod
    def find_wrapper(component):
        """Find the correct wrapper for this UIA component"""
        # Set a general wrapper by default
        wrapper_match = UIBaseComponent

        # Check for a more specific wrapper in the registry
        control_type = component.element_info.control_type

        if control_type in UIBaseComponentWrapperMeta.control_type_to_cls:
            wrapper_match = UIBaseComponentWrapperMeta.control_type_to_cls[control_type]
        else:
            raise NotImplementedError(f"{component} doesn't have an implemented wrapper")

        return wrapper_match






#___DEFINING_NATIVE_METHODS___

class UIBaseComponent(UIBaseComponentWrapper, metaclass=UIBaseComponentMeta):

    def __new__(cls, component: UIAWrapper):
        """Construct the control wrapper"""
        return super(UIBaseComponent, cls)._create_wrapper(cls, component, UIBaseComponent)

    # -----------------------------------------------------------
    def __init__(self, component: UIAWrapper):
        UIBaseComponentWrapper.__init__(self, component)

    def getValue(self):
        return self.component.iface_value
    
    def setValue(self, value):
        self.component.set_value(value)
    
    def setFocus(self):
        self.component.set_focus()
    
    def invoke(self):
        self.component.invoke()

    def click(self):
        self.component.click_input()
    
    def right_click(self):
        self.component.right_click_input()
    
    def isTopLevel(self):
        x, y, _, _= self.getCoordinates()
        return self.component.from_point(x, y).element_info.name == self.component.element_info.name

    def isVisible(self):
        return self.component.is_visible()
    
    def exists(self):
        raise self.isVisible()
    
    def getMidpoint(self):
        return self.component.rectangle().mid_point()
    
    def getCoordinates(self):
        rectangle = self.component.rectangle()
        return rectangle.left, rectangle.top, rectangle.right, rectangle.bottom
    
    def getSizes(self):
        rectangle = self.component.rectangle()
        return rectangle.width(), rectangle.height()
    
    def getChildren(self):
        return list(UIBaseComponent(component) for component in self.component.children())

    def getDescendants(self):
        return list(UIBaseComponent(component) for component in self.component.descendants())
    
    def _wait_find(self, function, check_function, control_type: Type, timeout: int = 1, retry_interval: float = 0.01, **criteria):
        timeafter = time.time() + timeout
        while time.time() < timeafter:
            item = function(control_type=control_type.native_control_type, **criteria)
            item_check = check_function(item)
            if item_check:
                if type(item) == list:
                    return list(UIBaseComponent(i) for i in item)
                else:
                    return UIBaseComponent(item)
            time.sleep(retry_interval)
        raise ElementNotFound(f"CheckFunction returned: Element - ControlType: {control_type.native_control_type} - {criteria} - not found after {timeout} seconds")
        
    def findAll(self, control_type: Type, timeout: int = 1, retry_interval: float = 0.01, **criteria):
        return self._wait_find(function=self.component.children,
                        check_function=self._check_elements_exist,
                        control_type=control_type,
                        timeout=timeout,
                        retry_interval=retry_interval,
                        **criteria)
    
    def findFirst(self, control_type: Type, timeout: int = 1, retry_interval: float = 0.01, **criteria):
        return self.findAll(control_type, timeout, retry_interval, **criteria)[0]

    def find(self, control_type: Type, timeout: int = 1, retry_interval: float = 0.01, **criteria):
        items = self.findAll(control_type, timeout, retry_interval, **criteria)
        
        if len(items) > 1:
            raise ElementNotFound(f"{len(items)} Elements found matching the criteria - ControlType: {control_type} - {criteria}")

        return items[0]
    
    def findAllR(self, control_type: Type, timeout: int = 1, retry_interval: float = 0.01, **criteria):
        return self._wait_find(function=self.component.descendants,
                        check_function=self._check_elements_exist,
                        control_type=control_type,
                        timeout=timeout,
                        retry_interval=retry_interval,
                        **criteria)
    
    def findFirstR(self, control_type: Type, timeout: int = 1, retry_interval: float = 0.01, **criteria):
        return self.findAllR(control_type, timeout, retry_interval, **criteria)[0]

    def findR(self, control_type: Type, timeout: int = 1, retry_interval: float = 0.01, **criteria):
        items = self.findAllR(control_type, timeout, retry_interval, **criteria)
        
        if len(items) > 1:
            raise ElementNotFound(f"{len(items)} Elements found matching the criteria - ControlType: {control_type} - {criteria}")

        return items[0]
    
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return None
    
    @classmethod
    @property
    def control_type(cls) -> Type:
        return type(cls)
    
    @property
    def title(self):
        return self.component.element_info.name
    
    

class UIButton(UIButtonWrapper, UIBaseComponent):
    native_control_type: str = "Button"

    def getValue(self):
        return self.component.get_toggle_state()
    
    def press(self):
        self.component.click()
    
    def setValue(self, value):
        return super().setValue(value)

class UIRadioButton(UIRadioButtonWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "RadioButton"

class UIText(UITextWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "Text"
    
class UISlider(UISliderWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "Slider"

    def getValue(self):
        return self.component.value()

class UIEdit(UIEditWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "Edit"

    def getValue(self):
        return self.component.get_value()
    
    def setValue(self, value):
        self.component.set_focus()
        self.component.invoke()
        send_keys(str(value) + "{ENTER}")

class UIMenu(UIMenuWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "Menu"
    
class UIMenuItem(UIMenuItemWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "MenuItem"

    def getValue(self):
        try:
            return bool(self.component.iface_toggle)
        except NoPatternInterfaceError:
            return False

    def select(self):
        self.component.select()
    
    def setValue(self, value):
        return super().setValue(value)

class UIWindow(UIWindowWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "Window"
    
    def moveResize(self, x=None, y=None, width=None, height=None):
        cur_rect = self.component.rectangle()

        # if no X is specified - so use current coordinate
        if x is None:
            x = cur_rect.left

        # if no Y is specified - so use current coordinate
        if y is None:
            y = cur_rect.top

        # if no width is specified - so use current width
        if width is None:
            width = cur_rect.width()

        # if no height is specified - so use current height
        if height is None:
            height = cur_rect.height()

        # ask for the window to be moved
        self.component.iface_transform.Move(x, y)
        self.component.iface_transform.Resize(width, height)

    def close(self):
        self.component.close()

class UIGroup(UIGroupWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "Custom"

class UIStaticText(UIStaticTextWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "Text"

    def getValue(self):
        return self.component.window_text()

class UITitleBar(UITitleBarWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "TitleBar"
        
class UIMenuBar(UIMenuBarWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "MenuBar"

class UIProgressBar(UIProgressBarWrapper, UIBaseComponent):
    @classmethod
    @property
    def native_control_type(cls) -> str:
        return "ProgressBar"

    def getValue(self):
        return int(self.component.legacy_properties()['Value'])

# !!! Mac Specific !!!

class UIMenuBarItem(UIMenuItem, UIBaseComponent): ...


from platform import system

if system() == "Darwin":
    from pyuiauto.mac.application import UIApplication

elif system() == "Windows":
    from pyuiauto.win.application import UIApplication

else:
    raise OSError("The current OS isn't supported with this framework")




# ___MODULES___
from __future__ import annotations
import time
from abc import ABC, abstractmethod
import logging
import contextlib
from typing import Union

try:
    from pyautogui import press
except ImportError: # requires pip install
        raise ModuleNotFoundError('To install the required modules use pip install pyautogui')

# __MY_MODULES__
from pyuiauto.components import UIWindow, UIButton, UIMenuItem, UIMenuBarItem
from pyuiauto.exceptions import ProcessNotFoundError, ElementNotFound


# ___CLASSES___

class UISystemTrayIconWrapper(ABC):
    '''UISystemTrayIconWrapper\n
    Provides a system tray icon manager to help find the system tray icon for an application\n
    whether it be hidden within the notification expand window or on the taskbar.'''
    def __init__(self, app: UIApplicationWrapper):
        self.app = app

    @abstractmethod
    def __enter__(self) -> UIButton:
        'Using a context manager helps to creates the UIButton dynamically.'

    @abstractmethod
    def __exit__(self, *args) -> None:
        'Using a context manager helps to dynamically tidy any open windows.'
    
    
class UIPopupMenuWrapper(ABC):
    '''UIPopupMenuWrapper\n
    Provides a popup manager for popup windows to help select specific menu item\n
    and provides a cleanup of popup windows.\n
    
    The popup_naming_scheme can vary from app to app however, these are some examples:\n
        - *the application name*\n
        - "Context"
        - "Pop-up"
    '''
    def __init__(self, app: UIApplicationWrapper, popup_naming_scheme: str = None) -> None:
        self.app = app
        self.win_name = popup_naming_scheme if popup_naming_scheme else app.appName
        self.current_popup = None
        self.steps = 0

    @abstractmethod
    def getMenuItemFromPath(self, *path: str) -> UIMenuItem:
        '''PopupMenu class get menu item from path method\n
        Uses the specified path to return a menu item component at [-1] index of the path.'''
    
    @abstractmethod
    def back(self):
        'Go back a popup menu window to the previous item'

    @abstractmethod
    def backToRoot(self):
        'Go back to the root popup menu window'
    
    @abstractmethod
    def __enter__(self) -> UIPopupMenuWrapper:
        'Using a context manager helps to creates the popup windows dynamically.'
    
    @abstractmethod
    def __exit__(self, *args) -> None:
        'Using a context manager helps to dynamically tidy any open windows.'


class UIApplicationWrapper(ABC):
    def __init__(self, appName: str, appPath: str = None) -> None:
        self.appName = appName
        self.appPath = appPath
        self.end_time = 0
        self.start_time = 0

    @abstractmethod
    def launchApp(self) -> None:
        '''Application class launch app method\n
        Launches the application.'''
        self.start_time = time.time()

    @abstractmethod
    def connectApp(self) -> None:
        '''Application class connect app method\n
        Connects the application wrapper to the specfied running app.'''

    @abstractmethod
    def terminateApp(self) -> None:
        '''Application class terminate app method\n
        Force quits the application.'''
        self.end_time = time.time()
    
    def getRuntime(self):
        return self.end_time - self.start_time
    
    @abstractmethod
    def window(self, timeout: int = 1, retry_interval: float = 0.01, **criteria) -> UIWindow:
        '''Application class window method\n
        Finds a window child component with the specified criteria.
        Raises an exception if the criteria matches no components (recursive).\n
        Args:                    
                    (criteria)
                    title: the name of the component (of type string)

                    (extra OS specific accessibility API criteria)
        Returns: 
                    component: window component wrapped to it's cross compatible custom wrapper of type UIWindowWrapper'''
    
    @abstractmethod
    def windows(self, timeout: int = 1, retry_interval: float = 0.01, **criteria) -> list[UIWindow]:
        '''Application class windows method\n
        Finds all window child components with the specified criteria.
        Raises an exception if the criteria matches > 1 or no components (recursive).\n
        Args:                    
                    (criteria)
                    title: the name of the component (of type string)

                    (extra OS specific accessibility API criteria)
        Returns: 
                    components: window components wrapped to it's cross compatible custom wrapper in a list of type UIWindowWrapper'''

    @abstractmethod
    def isAppAlreadyRunning(self) -> bool:
        '''Application class is app already running method\n
        This method checks if the application name is found in already running instances.'''

    @abstractmethod
    def isAppRunning(self) -> bool:
        '''Application class is app running method\n
        This method checks if the application is still running which can help to identify crashes.'''

    @abstractmethod
    def getSystemTrayIcon(self) -> UISystemTrayIconWrapper:
        '''Application class get system tray icon method\n
        Returns the system tray icon button component wrapper.\n
        (a context manager is required to interact with this component)\n
        Example:\n
        with UISystemTrayIcon() as icon:
            icon.right_click()
            ...'''
    
    @abstractmethod
    def getPopupMenu(self, popup_naming_scheme: str = None) -> UIPopupMenuWrapper:
        '''Application class get popup menu method\n
        Returns the popup menu component wrapper.\n
        (a context manager is required to interact with this component)\n
        Example:\n
        with UIPopupMenuWrapper() as popup:
            item = popup.getMenuItemFromPath("Setup", "Set Sample Rate", "96kHz")
            ...\n
        
        The popup_naming_scheme can vary from app to app however, these are some examples:\n
        - *the application name* (default)
        - "Context"
        - "Pop-up"
        '''

    @abstractmethod
    def systemTrayPopupPath(self, *path: str) -> UIMenuItem:
        '''Application class system tray popup select method\n
        This method selects menu items after clicking the system tray, from the *args passed in as a path.'''
    
    @abstractmethod
    def menuBarPopupPath(self, window: UIWindow, *path: str) -> Union[UIMenuBarItem, UIMenuItem]:
        '''Application class menu bar select method\n
        This method selects menu items on the menubar from the *args passed in as a path.'''

    @abstractmethod
    def getCrashReport(self) -> str:
        '''Application class get crash report method\n
        Gets all crash reports between the start and end time of the application running.'''

    def relaunchApp(self):
        '''Application class relaunch app method\n
        Uses it's context manager __enter__ __exit__ methods to quit and reopen the application.'''
        self.__exit__()
        self.launchApp()
        self.__enter__()

    def __enter__(self) -> UIApplicationWrapper:
        self.launchApp()
        self.connectApp()
        return self
    
    def __exit__(self, *args):
        if not self.isAppRunning():
            logging.warning(f"The {self.appName} application may have crashed as it was not found in running applications")
            # logging.error(f"Crash report:\n{self.getCrashReport()}")
            # raise ProcessNotFoundError(f"The {self.appName} application was not found in running applications")
        else:
            self.terminateApp()



#___MODULES___
from __future__ import annotations
from typing import Type
import os
import datetime
import logging
import contextlib
from typing import Union

# pip installed modules
try:
        import atomacos
except ImportError: # requires pip install
        raise ModuleNotFoundError('To install the required modules use pip install atomacos (MacOS ONLY)')

try:
    from pyautogui import press
except ImportError: # requires pip install
        raise ModuleNotFoundError('To install the required modules use pip install pyautogui')


#___MY_MODULES___
from pyuiauto.mac.components import UIBaseComponent, UIWindow, UIButton, UIMenuBarItem, UIMenuItem
from pyuiauto.base.application import UIApplicationWrapper, UIPopupMenuWrapper, UISystemTrayIconWrapper, UIPopupMenuWrapper
from pyuiauto.exceptions import ElementNotFound, WindowNotFound


class UISystemTrayIcon(UISystemTrayIconWrapper):
    def __init__(self, app: UIApplicationWrapper):
        super().__init__(app)

    def __enter__(self) -> UIMenuBarItem:
        return self.app._findFirstR(control_type=UIMenuBarItem, AXSubrole="AXMenuExtra")
    
    def __exit__(self, *args):
        pass

class UIPopupMenu(UIPopupMenuWrapper):
    def __init__(self, app: UIApplicationWrapper, popup_naming_scheme: str = None) -> None:
        super().__init__(app, popup_naming_scheme)

    def getMenuItemFromPath(self, *path: str) -> UIMenuItem:
        current_item = None

        for count, i in enumerate(path):
            self.current_popup = self.app.window(title=self.win_name)
            current_item = self.current_popup.findFirstR(title=i, control_type=UIMenuItem)
            if count != len(path)-1:
                current_item.invoke()
            self.steps += 1
        
        return current_item
    
    def _nativeMenuBarItemFromPath(self, *path: str) -> UIMenuBarItem:
        current_item = UIMenuItem(self.app._app.menuItem(*path))
        self.steps = len(path)
        
        return current_item
        
    def back(self):
        if self.steps > 0:
            press("left")
            self.steps -= 1

    def backToRoot(self):
        for i in range(self.steps):
            press("left")
            self.steps -= 1
    
    def __enter__(self) -> UIPopupMenu:
        return self
        
    def __exit__(self, *args) -> None:
        if self.current_popup != None and self.current_popup.exists():
            press("esc")


#___DEFINING_NATIVE_METHODS___

class UIApplication(UIApplicationWrapper):

    def launchApp(self) -> None:
        atomacos.launchAppByBundlePath(self.appPath)
    
    def connectApp(self) -> None:
        self._app = atomacos.getAppRefByLocalizedName(self.appName)
        self._app.activate()
        self._bundleID = self._app.getBundleId()
    
    def terminateApp(self) -> None:
        atomacos.terminateAppByBundleId(self._bundleID)
    
    def _findAll(self, control_type: Type[UIBaseComponent], **options):
        return UIBaseComponent(self._app).findAll(
                                                    control_type=control_type,
                                                    **options)
    
    def _findFirst(self, control_type: Type[UIBaseComponent], **options):
        return UIBaseComponent(self._app).findFirst(
                                                    control_type=control_type,
                                                    **options)
    
    def _findFirstR(self, control_type: Type[UIBaseComponent], **options):
        return UIBaseComponent(self._app).findFirstR(
                                                    control_type=control_type,
                                                    **options)
    
    def _findAllR(self, control_type: Type[UIBaseComponent], **options):
        return UIBaseComponent(self._app).findAllR(
                                                    control_type=control_type,
                                                    **options)
    
    def _find(self, control_type: Type[UIBaseComponent], **options):
        return UIBaseComponent(self._app).find(
                                                    control_type=control_type,
                                                    **options)

    def _findR(self, control_type: Type[UIBaseComponent], **options):
        return UIBaseComponent(self._app).findR(
                                                    control_type=control_type,
                                                    **options)

    def window(self, timeout: int = 1, retry_interval: float = 0.01, **options) -> UIWindow:
        try:
            return self.windows(timeout=timeout, retry_interval=retry_interval, **options)[-1]
        except ElementNotFound:
             raise WindowNotFound
    
    def windows(self, timeout: int = 1, retry_interval: float = 0.01, **options) -> list[UIWindow]:
        try:
            return self._findAllR(control_type=UIWindow, timeout=timeout, retry_interval=retry_interval, **options)
        except ElementNotFound:
             raise WindowNotFound
    
    def isAppAlreadyRunning(self):
        for i in atomacos.NativeUIElement.getRunningApps():
            if self.appPath in str(i.executableURL()):
                return True
        return False
    
    def isAppRunning(self):
        return atomacos.NativeUIElement._running_app
    
    def getSystemTrayIcon(self):
        return UISystemTrayIcon(self) # AXSubRole is MacOS ONLY criteria

    def getCrashReport(self):
        diagnostic_reports_path = os.path.join(os.path.expanduser('~'), "Library", "Logs", "DiagnosticReports")

        for file in os.listdir(diagnostic_reports_path):
            if file.endswith(".ips"):
                app, year, month, date, time_ = file.split("-")
                time_ = time_.split(".")[0]

                if app == self.appName:

                    file_time = datetime.datetime(int(year), int(month), int(date), int(time_[:2]), int(time_[2:4]), int(time_[4:]))
                    start_time = datetime.datetime.fromtimestamp(self.start_time)
                    end_time = datetime.datetime.fromtimestamp(self.end_time)
                    
                    logging.debug(f"Start Time: {start_time}, End Time: {end_time}, File Time: {file_time}")
                    
                    
                    if ( (start_time <= file_time) and (end_time >= file_time) ):
                        return os.path.join(diagnostic_reports_path, file)

    def getPopupMenu(self, popup_naming_scheme: str = None) -> UIPopupMenu:
        return UIPopupMenu(self, popup_naming_scheme)
    
    def getSystemTrayIcon(self) -> UISystemTrayIcon:
        return UISystemTrayIcon(self)
    
    @contextlib.contextmanager
    def systemTrayPopupPath(self, *path: str) -> UIMenuItem:
        with self.getSystemTrayIcon() as icon:
            icon.right_click()
        
            with self.getPopupMenu() as popup:
                try:

                    yield popup.getMenuItemFromPath(*path)
                except ElementNotFound:
                    raise ElementNotFound(f"{path} is disabled or not available")
    
    @contextlib.contextmanager
    def menuBarPopupPath(self, window: UIWindow, *path: str) -> Union[UIMenuBarItem, UIMenuItem]:
        try:
            menuBarItem = self._findFirstR(title=path[0], control_type=UIMenuBarItem)
            if len(path) == 1:
                yield menuBarItem
            
            with self.getPopupMenu(popup_naming_scheme="") as popup:
                yield popup._nativeMenuBarItemFromPath(*path)
        
        except ElementNotFound:
                    raise ElementNotFound(f"{path} is disabled or not available")